import jwt
import typing as t

from calendar import timegm
from aiosmtplib import send
from aiohttp import ClientSession
from email.mime.text import MIMEText
from os import environ as os_environ
from certifi import where as ssl_where
from datetime import datetime as dt, timedelta, UTC


from .exceptions import InternalError
from .result import Result, fail, ok, is_ok
from .domain import DecodedJwtToken, IPInfo
from .config import SMTPConfig, JwtConfig, SSL


__all__ = ["send_email", "decode_jwt_token", "create_jwt_token", "fetch_ip_info"]


async def send_email(to: str, title: str, html_body: str, config: SMTPConfig = SMTPConfig.DEFAULT) -> None:

    os_environ["SSL_CERT_FILE"] = ssl_where()
    sender = f"{config.sender_name} <{config.sender_email}>"
    message = MIMEText(html_body, "html")
    message["Subject"] = title
    message["From"] = sender
    message["To"] = to
    _ = await send(
        message, hostname=config.host, port=config.port, username=config.user, password=config.password, use_tls=True
    )


def decode_jwt_token(token: str, kind: str, config: JwtConfig = JwtConfig.DEFAULT) -> DecodedJwtToken:
    try:
        params: dict[str, t.Any] = {
            "jwt": token,
            "issuer": config.issuer,
            "key": config.decode_key,
            "audience": config.audience,
            "algorithms": [config.algorithm],
        }
        payload = jwt.decode(**params)
        kind, sub, extra = payload.get("kind", ""), payload.get("sub", ""), payload.get("extra", {})
        if kind != kind or not sub:
            return DecodedJwtToken(invalid=True)
        return DecodedJwtToken(sub=sub, kind=kind, extra=extra)
    except jwt.ExpiredSignatureError:
        return DecodedJwtToken(expired=True)
    except Exception:
        return DecodedJwtToken(invalid=True)


def create_jwt_token(
    sub: str, kind: str, duration_min: int, extra: dict[str, t.Any] | None = None, config: JwtConfig = JwtConfig.DEFAULT
) -> Result[str]:
    try:
        iat = dt.now(tz=UTC)
        exp = iat + timedelta(minutes=duration_min)
        payload = {
            "sub": sub,
            "kind": kind,
            "extra": extra or {},
            "iss": config.issuer,
            "aud": config.audience,
            "exp": timegm(exp.utctimetuple()),
            "iat": timegm(iat.utctimetuple()),
        }
        return ok(value=jwt.encode(payload=payload, key=config.encode_key, algorithm=config.algorithm))
    except Exception as e:
        return fail(err=InternalError.from_exception(e))


async def fetch_ip_info(ip: str) -> IPInfo | None:
    try:
        async with ClientSession() as session:
            async with session.get(f"https://ipinfo.io/{ip}/json", ssl=SSL) as resp:
                if resp.status != 200:
                    return None
                data = await resp.json()
                raw = {
                    "city": data.get("city"),
                    "country_code": data.get("country"),
                    "region_code": data.get("region_code"),
                    "timezone": data.get("timezone", "UTC"),
                    "currency": data.get("currency", "EUR"),
                }
                res = IPInfo.create(raw)
                return res.value if is_ok(res) else None
    except:
        return None



import typing as t
from attrs import define, field
from skoll.utils import get_config_var
from ssl import create_default_context
from certifi import where as ssl_where


__all__ = ["SMTPConfig", "JwtConfig", "SSL"]


@define(kw_only=True, slots=True, frozen=True)
class SMTPConfig:

    DEFAULT: t.ClassVar[t.Self]

    host: str = field(factory=get_config_var(keys=["SMTP_HOST"], default=""))
    port: int = field(factory=get_config_var(keys=["SMTP_PORT"], default=465))
    user: str = field(factory=get_config_var(keys=["SMTP_USER"], default=""))
    password: str = field(factory=get_config_var(keys=["SMTP_PASSWORD"], default=""))
    sender_name: str = field(factory=get_config_var(keys=["SMTP_SENDER_NAME"], default=""))
    sender_email: str = field(factory=get_config_var(keys=["SMTP_SENDER_EMAIL"], default=""))


@define(frozen=True, slots=True, kw_only=True)
class JwtConfig:

    DEFAULT: t.ClassVar[t.Self]

    issuer: str = field(factory=get_config_var(keys=["JWT_TOKEN_ISSUER"], default=""))
    encode_key: str = field(factory=get_config_var(keys=["JWT_HASH_SECRET"], default=""))
    decode_key: str = field(factory=get_config_var(keys=["JWT_HASH_SECRET"], default=""))
    audience: str = field(factory=get_config_var(keys=["JWT_TOKEN_AUDIENCE"], default=""))
    algorithm: str = field(factory=get_config_var(keys=["JWT_ALGORITHM"], default="HS256"))


JwtConfig.DEFAULT = JwtConfig()
SMTPConfig.DEFAULT = SMTPConfig()
SSL = create_default_context(cafile=ssl_where())



from .primitives import Enum

__all__ = ["SortDirection", "Status", "ThemeMode", "UnitSystem"]


class SortDirection(Enum):

    ASCENDING = "ASC"
    DESCENDING = "DESC"


class Status(Enum):

    ACTIVE = "ACTIVE"
    DELETED = "DELETED"
    INACTIVE = "INACTIVE"
    ARCHIVED = "ARCHIVED"


class ThemeMode(Enum):

    DARK = "DARK"
    LIGHT = "LIGHT"
    SYSTEM = "SYSTEM"


class UnitSystem(Enum):

    METRIC = "METRIC"
    IMPERIAL = "IMPERIAL"
    US_CUSTOMARY = "US_CUSTOMARY"



import typing as t
import collections.abc as c

from attrs import define, field
from skoll.result import is_ok, Result
from inspect import iscoroutinefunction, signature


from .primitives import Object, ID, DateTime, Locale, Map, Timezone

__all__ = [
    "Service",
    "Message",
    "MsgClient",
    "RawMessage",
    "MsgContext",
    "Subscriber",
    "RawMsgClient",
    "RawMsgContext",
    "MsgClientDevice",
    "RawMsgClientDevice",
    "SubscriberCallback",
]

type SubscriberCallback[T: Message] = t.Callable[t.Concatenate[T, ...], c.Coroutine[t.Any, t.Any, Result[t.Any]]]


class RawMsgClientDevice(t.TypedDict):
    duid: t.NotRequired[str]
    model: t.NotRequired[str]
    brand: t.NotRequired[str]
    os_name: t.NotRequired[str]
    os_version: t.NotRequired[str]


class RawMsgClient(t.TypedDict):
    ip_address: t.NotRequired[str]
    locale: t.NotRequired[str]
    timezone: t.NotRequired[str]
    device: t.NotRequired[RawMsgClientDevice]


class RawMsgContext(t.TypedDict):
    user_id: t.NotRequired[str]
    span_id: t.NotRequired[str]
    trace_id: t.NotRequired[str]
    client: t.NotRequired[RawMsgClient]
    extra: t.NotRequired[dict[str, t.Any]]


class RawMessage(t.TypedDict):
    name: str
    source: str
    id: t.NotRequired[str]
    created_at: t.NotRequired[int]
    context: t.NotRequired[RawMsgContext]
    payload: t.NotRequired[dict[str, t.Any]]


@define(frozen=True, kw_only=True, slots=True)
class MsgClientDevice(Object):

    duid: ID | None = None
    model: str | None = None
    brand: str | None = None
    os_name: str | None = None
    os_version: str | None = None

    @classmethod
    def default(cls) -> t.Self:
        return cls()


@define(frozen=True, kw_only=True, slots=True)
class MsgClient(Object):

    ip_address: str | None = None
    locale: Locale = field(factory=Locale.default)
    timezone: Timezone = field(factory=Timezone.default)
    device: MsgClientDevice = field(factory=MsgClientDevice.default)

    @classmethod
    def default(cls) -> t.Self:
        return cls()


@define(frozen=True, kw_only=True, slots=True)
class MsgContext(Object):

    user_id: ID | None = None
    span_id: ID = field(factory=ID.new)
    trace_id: ID = field(factory=ID.new)
    extra: Map = field(factory=Map.empty)
    client: MsgClient = field(factory=MsgClient.default)

    @classmethod
    def default(cls) -> t.Self:
        return cls()


@define(frozen=True, kw_only=True, slots=True, eq=False)
class Message(Object):

    name: str
    source: str
    payload: Object
    id: ID = field(factory=ID.new)
    created_at: DateTime = field(factory=DateTime.now)
    context: MsgContext = field(factory=MsgContext.default)

    @t.override
    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return other.__hash__() == self.__hash__()

    @t.override
    def __ne__(self, other: t.Any) -> bool:
        return not self == other

    @t.override
    def __hash__(self) -> int:
        return hash(self.id.serialize())

    @classmethod
    def from_raw(cls, raw: RawMessage) -> t.Self | None:
        res = cls.create(raw)
        if is_ok(res):
            return res.value
        return None


@define(frozen=True, kw_only=True, slots=True)
class Subscriber[T: Message]:

    topic: str
    msg_arg: str
    queued: bool
    will_reply: bool
    service_name: str
    msg_type: type[Message]
    js_stream: str | None = None
    callback: SubscriberCallback[T]


@define(kw_only=True, slots=True, frozen=True)
class Service:

    name: str
    subscribers: list[Subscriber[t.Any]] = field(factory=list)

    def _add[T: Message](
        self, topic: str, will_reply: bool, queued: bool, callback: SubscriberCallback[T], js_stream: str | None = None
    ):
        first_arg = list(signature(callback).parameters.values())[0]
        if not issubclass(first_arg.annotation, Message) or not iscoroutinefunction(callback):
            raise TypeError(
                f"Service subscriber @on/@reply must be a coroutine and with first argument being a subclass of Message"
            )
        self.subscribers.append(
            Subscriber(
                topic=topic,
                queued=queued,
                callback=callback,
                js_stream=js_stream,
                will_reply=will_reply,
                msg_arg=first_arg.name,
                service_name=self.name,
                msg_type=first_arg.annotation,
            )
        )

    def on(self, topic: str, queued: bool = False, stream: str | None = None):
        def decorator[T: Message](callback: SubscriberCallback[T]):
            self._add(topic, will_reply=False, queued=queued, callback=callback, js_stream=stream)
            return callback

        return decorator

    def reply(self, topic: str):
        def decorator[T: Message](callback: SubscriberCallback[T]):
            self._add(topic, will_reply=True, queued=True, callback=callback)
            return callback

        return decorator



import typing as t

from skoll.result import Result
from skoll.config import SMTPConfig

from .primitives import ID
from .objects import Entity, IPInfo
from .messaging import Message, RawMessage, Service
from .typing import Criteria, ListPage, AuthzWriteChange, AuthzPrecondition, AuthzLookupResult


__all__ = ["DB", "Repository", "Mediator", "Authz", "EmailSender", "IPInfoProvider"]


class IPInfoProvider(t.Protocol):

    async def __call__(self, ip: str) -> IPInfo | None: ...


class DB[T = t.Any](t.Protocol):

    async def close(self) -> None: ...
    async def connect(self) -> None: ...
    def session(self) -> t.AsyncContextManager[T]: ...
    def transaction(self) -> t.AsyncContextManager[T]: ...


class Repository[T: Entity](t.Protocol):

    async def save(self, state: T) -> None: ...
    async def delete(self, criteria: Criteria) -> None: ...
    async def exist(self, criteria: Criteria) -> bool: ...
    async def get(self, criteria: Criteria) -> T | None: ...
    async def list(self, criteria: Criteria) -> ListPage[T]: ...


class Mediator(t.Protocol):

    async def connect(self) -> None: ...
    async def disconnect(self) -> None: ...
    async def unsubscribe(self, id: ID) -> None: ...

    async def subscribe(self, service: Service) -> ID: ...
    async def publish(self, *msg: Message | RawMessage) -> None: ...
    async def request(self, msg: Message | RawMessage) -> Result[t.Any]: ...


class Authz(t.Protocol):

    async def check(self, tuple: str, cxt: dict[str, t.Any] | None = None) -> None: ...
    async def write(
        self, changes: list[AuthzWriteChange], preconditions: list[AuthzPrecondition] | None = None
    ) -> str: ...
    async def lookup(
        self, filter: str, cxt: dict[str, t.Any] | None = None, limit: int | None = None, cursor: str | None = None
    ) -> AuthzLookupResult: ...


class EmailSender(t.Protocol):

    async def __call__(self, to: str, title: str, html_body: str, config: SMTPConfig = SMTPConfig.DEFAULT) -> None: ...



import typing as t

from attrs import define, field
from abc import ABC, abstractmethod

from .primitives import ID, DateTime


__all__ = [
    "Criteria",
    "ListPage",
    "SQLCriteria",
    "DecodedJwtToken",
    "AuthzWriteChange",
    "AuthzPrecondition",
    "AuthzLookupResult",
    "AuthzWriteOperation",
    "AuthzPreconditionOperation",
]

type AuthzWriteOperation = t.Literal["SET", "DELETE"]
type AuthzPrecondition = tuple[AuthzPreconditionOperation, str]
type AuthzPreconditionOperation = t.Literal["MUST_MATCH", "MUST_NOT_MATCH"]
type AuthzWriteChange = tuple[AuthzWriteOperation, str, tuple[str, dict[str, t.Any]] | None, DateTime | None]


class ListPage[T](t.NamedTuple):

    items: list[T]
    cursor: str | None = None


class SQLCriteria(t.NamedTuple):
    query: str
    params: list[t.Any]


class DecodedJwtToken(t.NamedTuple):

    expired: bool = False
    invalid: bool = False
    sub: str | None = None
    kind: str | None = None
    extra: dict[str, t.Any] | None = None


@define(frozen=True, kw_only=True, slots=True)
class Criteria(ABC):

    id: ID | None = None
    cursor: str | None = None
    limit: int = field(default=150)

    @property
    @abstractmethod
    def as_sql(self) -> SQLCriteria:
        raise NotImplementedError("Subclasses must implement this method")


class AuthzLookupResult(t.NamedTuple):
    ids: list[str]
    cursor: str | None = None



from .i18n import *
from .nats import *
from .utils import *
from .domain import *
from .result import *
from .helpers import *
from .spicedb import *
from .exceptions import *
from .postgresql import *



import json
import typing as t
import collections.abc as c

from nats.aio.msg import Msg
from attrs import define, field
from nats.js import JetStreamContext
from nats.aio.client import Client as NatsClient
from nats.aio.subscription import Subscription as NSubscription

from .config import SSL
from .utils import call_with_dependencies
from .result import Result, fail, ok, is_ok, is_fail
from .exceptions import Error, InternalError, ValidationFailed
from .domain import Message, ID, RawMessage, Subscriber, Service, Mediator

__all__ = ["NatsMediator"]


@define(kw_only=True, slots=True)
class NatsMediator(Mediator):

    creds: str
    servers: list[str]
    _js: JetStreamContext | None = None
    nc: NatsClient = field(init=False, factory=lambda: NatsClient())
    subscriptions: dict[str, list[NSubscription]] = field(factory=dict)

    @property
    def js(self) -> JetStreamContext:
        if self._js is None:
            raise InternalError(debug={"message": "JetStream context is not initialized"})
        return self._js

    @t.override
    async def subscribe(self, service: Service) -> ID:
        if not self.nc.is_connected:
            raise InternalError(debug={"message": "Attempt to subscribe before nats client is connected"})
        sub_id = ID.new()
        subscribtions: list[NSubscription] = []
        for subscriber in service.subscribers:
            if subscriber.js_stream is not None:
                sub = await self.js.subscribe(
                    manual_ack=True,
                    subject=subscriber.topic,
                    stream=subscriber.js_stream,
                    cb=wrap_callback(subscriber),
                    durable=subscriber.service_name,
                    queue=subscriber.service_name if subscriber.queued else None,
                )
            else:
                sub = await self.nc.subscribe(
                    subject=subscriber.topic,
                    cb=wrap_callback(subscriber),
                    queue=subscriber.service_name if subscriber.queued else "",
                )
            subscribtions.append(sub)
        self.subscriptions[sub_id.value] = subscribtions
        return sub_id

    @t.override
    async def unsubscribe(self, id: ID) -> None:
        try:
            subscriptions = self.subscriptions.get(id.value, [])
            for sub in subscriptions:
                await sub.unsubscribe(limit=0)
            del self.subscriptions[id.value]
        except Exception as e:
            raise InternalError.from_exception(e)

    @t.override
    async def connect(self) -> None:
        try:
            if self.nc.is_connected or self.nc.is_reconnecting:
                return None
            await self.nc.connect(tls=SSL, servers=self.servers, max_reconnect_attempts=-1, user_credentials=self.creds)
            self._js = self.nc.jetstream()
            return None
        except Exception as e:
            raise InternalError.from_exception(e)

    @t.override
    async def disconnect(self) -> None:
        if self.nc.is_connected:
            await self.nc.drain()

    @t.override
    async def publish(self, *msg: Message | RawMessage) -> None:
        for m in msg:
            subject, payload = (m.name, m.serialize()) if isinstance(m, Message) else (m["name"], m)
            await self.nc.publish(subject, json.dumps(payload).encode("utf-8"))

    @t.override
    async def request(self, msg: Message | RawMessage) -> Result[t.Any]:
        try:
            subject, payload = (msg.name, msg.serialize()) if isinstance(msg, Message) else (msg["name"], msg)
            response = await self.nc.request(subject, json.dumps(payload).encode("utf-8"), timeout=5)
            raw_msg = json.loads(response.data.decode("utf-8"))
            if raw_msg.get("error") is not None:
                return fail(Error.from_dict(raw_msg["error"]))
            return ok(raw_msg.get("data"))
        except TimeoutError as e:
            return fail(InternalError.from_exception(e, extra={"message": f"Request timed out"}))
        except Exception as e:
            return fail(InternalError.from_exception(e))


def get_message(cls: type[Message], message: Message | RawMessage) -> Message:
    if isinstance(message, Message):
        return message
    res = cls.create(raw=message)
    if is_fail(res):
        raise ValidationFailed(errors=res.err.errors)

    return res.value


async def run_callback(subscriber: Subscriber[t.Any], message: Message | RawMessage) -> Result[t.Any]:
    topic = message.name if isinstance(message, Message) else message.get("name")
    try:
        msg = get_message(subscriber.msg_type, message)
        return await call_with_dependencies(subscriber.callback, {subscriber.msg_arg: msg})
    except Error as err:
        return fail(err=err)
    except Exception as exc:
        return fail(err=InternalError.from_exception(exc, extra={"subject": topic, "message": message}))


def wrap_callback(subscriber: Subscriber[t.Any]) -> t.Callable[[Msg], c.Awaitable[None]]:
    async def callback(msg: Msg):
        try:
            raw_msg: RawMessage = json.loads(msg.data.decode("utf-8"))
            result = await run_callback(subscriber, raw_msg)
            if subscriber.will_reply:
                raw_response = {
                    "data": result.value if is_ok(result) else None,
                    "error": result.err.serialize() if not is_ok(result) else None,
                }
                await msg.respond(json.dumps(raw_response).encode("utf-8"))
            elif subscriber.js_stream is not None:
                if is_ok(result):
                    await msg.ack()
                else:
                    await msg.nak(delay=5)
        except Exception as e:
            print(InternalError.from_exception(e))

    return callback
